#include "cdreader.h"

#ifdef _WINDOWS
#include <mmreg.h>
#endif

//  cdda.c
//  play and stop funcs

// Special thanks to GameDev.net for the tutorial on winmm =D
// http://www.gamedev.net/reference/articles/article1348.asp

// my variables for CDDA playing
#define CDDA_BUFFER_SIZE 150*2352
unsigned char playFrom[3];
long currPlayPos;
long startPlayPos;
long bufferPlayPos;
long stopPlayPos;
long CDDABuffer[CDDA_BUFFER_SIZE/4];
int volLevel = 5;

// this is my callback for buffering the sound. pBuffer is the buffer where the
// sound goes, bufferLen is the number of bytes to fill
void bufferSound (void *pBuffer, long bufferLen)
{
   int i = 0;
   long* lBuffer = pBuffer;
   short* sBuffer = pBuffer;

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "fweh %d\n", currPlayPos);
   fclose(log);
#endif
   
      // for rar images      
   if (CD.fullBuffer)
   {
         // if you're going past the stop point, copy up to that point
         // then fill the rest of the buffer
      if ( (currPlayPos + bufferLen) >= stopPlayPos)
      {
         memcpy(lBuffer, CD.fullBuffer + currPlayPos, stopPlayPos - currPlayPos);
         lBuffer += stopPlayPos - currPlayPos;
         memcpy(lBuffer, CD.fullBuffer + startPlayPos, bufferLen - (stopPlayPos - currPlayPos));
         currPlayPos = startPlayPos + bufferLen - (stopPlayPos - currPlayPos);
      }
         // just fill the buffer
      else
      {
         memcpy(pBuffer,CD.fullBuffer + currPlayPos,bufferLen);
         currPlayPos += bufferLen;
      }
         // volume adjust
      for (i = 0; i < (bufferLen/4); i++)
      {
         *sBuffer++ /= volLevel;
         *sBuffer++ /= volLevel;
      }
   }
      // rebuffer when you've reached the end of the buffer (go fig)
   else 
   {
      for (i = 0; i < (bufferLen/4); i++)
      {
            // at end of track, rewind
         if (currPlayPos == stopPlayPos)
         {
            currPlayPos = startPlayPos;
            fseek(CD.cdda, currPlayPos, SEEK_SET);
            fread(CDDABuffer,sizeof(unsigned char),CDDA_BUFFER_SIZE, CD.cdda);
            bufferPlayPos = currPlayPos;
         }
            // if you've skipped to somewhere else or ran to the end
            // of the buffer
         else if ( (currPlayPos >= (bufferPlayPos + CDDA_BUFFER_SIZE)) ||
              (currPlayPos < bufferPlayPos) )
         {
            fseek(CD.cdda, currPlayPos, SEEK_SET);
            fread(CDDABuffer,sizeof(unsigned char),CDDA_BUFFER_SIZE, CD.cdda);
            bufferPlayPos = currPlayPos;
         }
         *lBuffer++ = CDDABuffer[(currPlayPos - bufferPlayPos)/4];
            // a sneaky way of adjusting the volume ;)
         *sBuffer++ /= volLevel;
         *sBuffer++ /= volLevel;
         currPlayPos += 4;
      }
   }
}

#ifdef _WINDOWS

// all the variables needed to run the winmm sound stuff

#define	REPLAY_RATE				44100
#define	REPLAY_DEPTH			16
#define	REPLAY_SAMPLELEN		(REPLAY_DEPTH/8)
#define	REPLAY_NBSOUNDBUFFER	2

BOOL	m_bServerRunning=FALSE;
HWND	m_hWnd;
long	m_bufferSize;
long	m_currentBuffer=0;
HWAVEOUT		m_hWaveOut;
WAVEHDR			m_waveHeader[REPLAY_NBSOUNDBUFFER];
void			*m_pSoundBuffer[REPLAY_NBSOUNDBUFFER];

// preps the sound buffers
void	fillNextBuffer(void)
{
		// check if the buffer is already prepared (should not !)
		if (m_waveHeader[m_currentBuffer].dwFlags&WHDR_PREPARED)
			waveOutUnprepareHeader(m_hWaveOut,&m_waveHeader[m_currentBuffer],sizeof(WAVEHDR));

		// Call the user function to fill the buffer with anything you want ! :-)
		
	   bufferSound(m_pSoundBuffer[m_currentBuffer],m_bufferSize);

		// Prepare the buffer to be sent to the WaveOut API
		m_waveHeader[m_currentBuffer].lpData = (char*)m_pSoundBuffer[m_currentBuffer];
		m_waveHeader[m_currentBuffer].dwBufferLength = m_bufferSize;
		waveOutPrepareHeader(m_hWaveOut,&m_waveHeader[m_currentBuffer],sizeof(WAVEHDR));

		// Send the buffer the the WaveOut queue
		waveOutWrite(m_hWaveOut,&m_waveHeader[m_currentBuffer],sizeof(WAVEHDR));

		m_currentBuffer++;
		if (m_currentBuffer >= REPLAY_NBSOUNDBUFFER) m_currentBuffer = 0;
}
#endif //_WINDOWS


#ifdef _WINDOWS

// Internal WaveOut API callback function. We just call our sound handler ("playNextBuffer")
void CALLBACK waveOutProc(HWAVEOUT hwo,UINT uMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2)
{
   if (!playing) return;

   if (uMsg == WOM_DONE)
	{
		fillNextBuffer();
	}
}

#endif _WINDOWS


// initializes the sound buffers
BOOL	CDDAopen(void)
{
#ifdef _WINDOWS
      WAVEFORMATEX wfx;
      MMRESULT errCode;
      int i;

      // we're buffering 1 second here
		m_bufferSize = REPLAY_RATE * REPLAY_SAMPLELEN * 2;
		m_bufferSize /= REPLAY_NBSOUNDBUFFER;
		
		wfx.wFormatTag = WAVE_FORMAT_PCM;
		wfx.nChannels = 2;
		wfx.nSamplesPerSec = REPLAY_RATE;
		wfx.nAvgBytesPerSec = REPLAY_RATE*REPLAY_SAMPLELEN;
		wfx.nBlockAlign = REPLAY_SAMPLELEN;
		wfx.wBitsPerSample = REPLAY_DEPTH;
		wfx.cbSize = 0;
		errCode = waveOutOpen(	&m_hWaveOut,
										WAVE_MAPPER,
										&wfx,
										(DWORD)waveOutProc,
										(DWORD)NULL,
										(DWORD)CALLBACK_FUNCTION);

      if (errCode != MMSYSERR_NOERROR)
      {
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "errcode %d\n", errCode);
   fclose(log);
#endif
         return FALSE;
      }

		// Alloc the sample buffers.
		for (i=0;i<REPLAY_NBSOUNDBUFFER;i++)
		{
			m_pSoundBuffer[i] = malloc(m_bufferSize);
			memset(&m_waveHeader[i],0,sizeof(WAVEHDR));
		}

		// Fill all the sound buffers
		m_currentBuffer = 0;
		for (i=0;i<REPLAY_NBSOUNDBUFFER;i++)
			fillNextBuffer();

		m_bServerRunning = TRUE;
		return TRUE;
#endif // _WINDOWS
}

// stops the CDDA stream by closing all the buffers
void	CDDAclose(void)
{
#ifdef _WINDOWS
      int i;
		if (m_bServerRunning)
		{
			waveOutReset(m_hWaveOut);
			for (i=0;i<REPLAY_NBSOUNDBUFFER;i++)
			{
				if (m_waveHeader[m_currentBuffer].dwFlags & WHDR_PREPARED)
            {
					waveOutUnprepareHeader(m_hWaveOut,&m_waveHeader[i],sizeof(WAVEHDR));
            }

				free(m_pSoundBuffer[i]);
			}
			waveOutClose(m_hWaveOut);
         m_bServerRunning = FALSE;
		}
#endif // _WINDOWS  
}

long stop(void)
{
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log,"stop\n");
   fclose(log);
#endif
   if (!playing) return 0;
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log,"stop while playing\n");
   fclose(log);
#endif   
   playing = FALSE;
   CDDAclose();
   return 0;
}

// plays CDDA data from image
long play(unsigned char m, unsigned char s, unsigned char f)
{
   int i;

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "play request\n");
   fclose(log);
#endif

   // only supported for cue sheets
   if (CD.type != Cue) return 0;

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "play0 %02d:%02d:%02d\n", (int)m, (int)s, (int)f);
   fclose(log);
#endif
      // ignore repeated requests to play the same track
   if ((playing) && (m == playFrom[0]) && (s == playFrom[1]) && (f = playFrom[2]))
   {
      return 0;
   }
      // sometimes someone calls play with a new track but doesn't stop it first =/
   else if (playing)
   {
      stop();
   }

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "play %02d:%02d:%02d\n", (int)m, (int)s, (int)f);
   fclose(log);
#endif

      // mark start time and play position and buffer data
   playing = TRUE;

   playFrom[0]=m;
   playFrom[1]=s;
   playFrom[2]=f;

   // the '-300*2352' is because all of my cd start times are
   // off 4 seconds because of offsets/gaps on the cd.  the
   // times are accurate to the actual cd, just not to what's
   // on the disk image
   startPlayPos = (((m * 60) + s) * 75 + f) * 2352 - 300 * 2352;
   currPlayPos = startPlayPos;
   bufferPlayPos = startPlayPos;
   stopPlayPos = -1;

      // find the track number for this track by subtracting the track
      // times.. if it's less than 2 sec difference, then you've
      // found the track
   for (i = 1; i < CD.numtracks; i++)
   {
      char temp[3], answer[3];
      temp[0] = m;
      temp[1] = s;
      temp[2] = f;
      sub(CD.tl[i].time,temp, answer);
#ifdef MOOBY_DEBUG
      log = fopen(LOGFILE,"a");
      fprintf(log, "answer %d %d %d\n",answer[0], answer[1], answer[2]);
      fclose(log);
#endif
         // it's this track.  the stop time is the start time of the next track
         // minus 4 seconds
      if ( (answer[0] == 0) && (answer[1] <= 2) )
      {
         if (i == CD.numtracks)
         {
            i = 0;
         }
         else
         {
            i++;
         }
         stopPlayPos = (((CD.tl[i].time[0] * 60) + CD.tl[i].time[1]) * 75 + 
            CD.tl[i].time[2]) * 2352 - 300*2352;
#ifdef MOOBY_DEBUG
         log = fopen(LOGFILE,"a");
         fprintf(log, "stopping at track %d\n", stopPlayPos);
         fclose(log);
#endif

         break;
      }
   }

      // didn't find the end...  just set it to the end of the CD
   if (stopPlayPos == -1)
   {
#ifdef MOOBY_DEBUG
         log = fopen(LOGFILE,"a");
         fprintf(log, "can't stop!\n");
         fclose(log);
#endif
      stopPlayPos = (((CD.tl[0].time[0] * 60) + CD.tl[0].time[1]) * 75 + 
         CD.tl[0].time[2]) * 2352;
   }

   if (CD.cdda)
   {
      fseek(CD.cdda, currPlayPos, SEEK_SET);
      fread(CDDABuffer,sizeof(unsigned char),2352, CD.cdda);
   }

   if (CDDAopen() != TRUE)
   {
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "CDDA toast\n");
   fclose(log);
#endif
   }
   return 0;
}

int CD_Play(unsigned char *param)
{
   return play(BCDToInt(param[0]), BCDToInt(param[1]), BCDToInt(param[2]));
}

long CALLBACK CDRplay(unsigned char * sector)
{
   if (CD.bcdFix)
   {
      return play(sector[0], sector[1], sector[2]);
   }
   else
   {
      return play(BCDToInt(sector[0]), BCDToInt(sector[1]), BCDToInt(sector[2]));
   }
}



int CD_Stop(void)
{
	return stop();
}


long CALLBACK CDRstop(void)
{
	return stop();
}
