#include "cdreader.h"

//  info.c
//  cd information like number of tracks and track length

struct CdrStat {
    unsigned long Type;
    unsigned long Status;
    unsigned char Time[3]; // current playing time
};

// reads cdr status
// type:
// 	0x00 - unknown
//	0x01 - data
//	0x02 - audio
//	0xff - no cdrom
// status:
//	0x00 - unknown
//	0x01 - error
//	0x04 - seek error
//	0x10 - shell open
//	0x20 - reading
//	0x40 - seeking
//	0x80 - playing
// time:
//	byte 0 - minute
//	byte 1 - second
//	byte 2 - frame
long CALLBACK CDRgetStatus(struct CdrStat *stat) 
{
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "cdrgetstatus\n");
   fclose(log);
#endif

  stat->Time[0] = 0;
  stat->Time[0] = 0;
  stat->Time[0] = 0;

  if (playing)
  {
     stat->Status=0x80;
     stat->Type = 0x02;
  }
  else
  {
     stat->Status=0x20;
     stat->Type = 0x01;
  }
  return 0;
}

char CALLBACK CDRgetDriveLetter(void)
{
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "getdriveletter\n");
   fclose(log);
#endif

   return 0;
}

// gets track 
long getTN(unsigned char* buffer)
{
 int numtracks = getNumTracks();

 if (-1 == numtracks)
 {

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "getnumtracks failed\n");
   fclose(log);
#endif

	return -1;
 }

 buffer[0]=1;
 buffer[1]=numtracks;

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "getnumtracks %d %d\n", (int)buffer[0], (int)buffer[1]);
   fclose(log);
#endif

   return 0;
}

// psemu callback
long CALLBACK CDRgetTN(unsigned char *buffer)
{
	return getTN(buffer);
   
}

// fpse callback
int CD_GetTN(unsigned char *result)
{
	return getTN(result + 1);
}


 // if track==0 -> return total length of cd
 // otherwise return start in bcd time format
long getTD(int track, unsigned char* buffer)
{
      // lasttrack just keeps track of which track TD was requested last (go fig)
   static int lasttrack = 1;

      // we're checking here whether or not the track requests are in BCD format
      // or not.  If they are, then this is FPSE calling the PSEmu interface
      // and we need to return the TDs in a different order.  I'm not really happy
      // with this code, so I'm gonna talk to the FPSE peepz hoping that they'll
      // correct the problem
   if ( (lasttrack == 9) && (track == 16) )
   {
#ifdef MOOBY_DEBUG
      log = fopen(LOGFILE,"a");
      fprintf(log, "bcdfix\n");
      fclose(log);
#endif
      CD.bcdFix = TRUE;
   }

   lasttrack = track;

   if (CD.bcdFix)
   {
      track = BCDToInt(track);
   }

   if (track > CD.numtracks)
   {
#ifdef MOOBY_DEBUG
      log = fopen(LOGFILE,"a");
      fprintf(log, "getTD bad %2d\n", track);
      fclose(log);
#endif
      return -1;
   }

   buffer[0] = CD.tl[track].time[0];
   buffer[1] = CD.tl[track].time[1];
   buffer[2] = CD.tl[track].time[2];
   

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "getTD %2d %02d:%02d:%02d\n", track, (int)buffer[0],
         (int)buffer[1], (int)buffer[2]);
   fclose(log);
#endif

   // bcd encode it
   buffer[0] = intToBCD(buffer[0]);
   buffer[1] = intToBCD(buffer[1]);
   buffer[2] = intToBCD(buffer[2]);
   
   return 0;

}

// psemu callback
long CALLBACK CDRgetTD(unsigned char track, unsigned char *buffer)
{
   unsigned char temp[3];
   int result = getTD((int)track, temp);

   if (result == -1) return -1;

      // once again, this is the fix for FPSE calling the PSEmu interface
   if (CD.bcdFix)
   {
      buffer[0] = temp[0];
      buffer[1] = temp[1];
      buffer[2] = temp[2];
   }
   else
   {
      buffer[0] = temp[2];
      buffer[1] = temp[1];
      buffer[2] = temp[0];
   }
   return 0;
}

// fpse callback
int CD_GetTD(unsigned char *result,int track)
{
   return getTD((int)(BCDToInt(track)), result + 1);
}


unsigned char * CALLBACK CDRgetBufferSub(void)
{
   return CD.subchannelPointer;
}