#include "cdreader.h"

#include "zlib/zlib.h"
// llinfo.c

// home of low level info and reading functions

// returns the number of tracks
char getNumTracks()
{
   // if there's no open cd, return -1
   if ( (!CD.cd) && (!CD.fullBuffer) )
   {
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "numtracks error\n");
   fclose(log);
#endif
      return -1;
   }

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "numtracks %d\n",CD.numtracks);
   fclose(log);
#endif
   return CD.numtracks;
}

void seekSector(const unsigned char m, const unsigned char s, const unsigned char f)
{
     // calc byte to search for
   CD.sector = (( (m * 60) + 
          (s - 2)) * 75 + f) * 2352;


   CD.subchannelPointer = &CD.subchannelData[(((m * 60) + s) * 75 + f) * BYTES_PER_SUBCHANNEL_FRAME];

#ifdef SUPER_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "seek %d %02d:%02d:%02d\n",CD.sector,
      (int)m,
      (int)s,
      (int)f);
   fclose(log);
#endif

   if (CD.fullBuffer)
   {
      return;
   }


   // read from file
   // is it cached?
   if (CD.type == IndexBZ)
   {
      if (! ((CD.sector >= CD.bufferPos) &&
          (CD.sector < (CD.bufferPos + BZIP_BUFFER_SECTORS * 2352))))
      {
         readit(m,s,f);
      }
   }
   else if (CD.type == IndexZ)
   {
      readit(m,s,f);
   }
   else if ((CD.sector >= CD.bufferPos) && 
       (CD.sector < (CD.bufferPos + BUFFER_SIZE)) )
   {
#ifdef SUPER_DEBUG
      log = fopen(LOGFILE,"a");
      fprintf(log, " cached %d %d\n",CD.sector - CD.bufferPos,BUFFER_SIZE);
      fclose(log);
#endif

      return;
   }

   // not cached - read a few blocks into the cache
   else
   {
      readit(m,s,f);
   }
}

void readit(const unsigned char m, const unsigned char s, const unsigned char f)
{
#ifdef SUPER_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " not cached %d %d\n",CD.sector - CD.bufferPos,BUFFER_SIZE);
   fclose(log);
#endif


   if (CD.type == IndexBZ)
   {
      unsigned int seekStart = CD.indexArray[((s-2+m*60)*75+f)/BZIP_BUFFER_SECTORS];
      unsigned int seekEnd = CD.indexArray[((s-2+m*60)*75+f)/BZIP_BUFFER_SECTORS+1];
      unsigned int destLen = BZIP_BUFFER_SECTORS * 2352;
      char tempBuffer[(BZIP_BUFFER_SECTORS + 1) * 2352];

      rc = fseek(CD.cd, seekStart, SEEK_SET);
#ifdef SUPER_DEBUG
      if (rc)
      {
   log = fopen(LOGFILE,"a");
   fprintf(log, " fseek %d rc %d\n",seekStart,rc);
   fclose(log);
      }
#endif

      rc = fread(tempBuffer, 1, seekEnd - seekStart, CD.cd);
#ifdef SUPER_DEBUG
      if (rc <= 0)
      {
   log = fopen(LOGFILE,"a");
   fprintf(log, " fread %d rc %d\n",seekEnd - seekStart,rc);
   fclose(log);
      }
#endif


       if ( (rc = BZ2_bzBuffToBuffDecompress ( CD.buffer,
                                    &destLen,
                                    tempBuffer,
                                    seekEnd - seekStart,
                                    0,
                                    0) ) != BZ_OK)

       {

#ifdef SUPER_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " error %d rc %d\n",CD.sector - CD.bufferPos,rc);
   fclose(log);
#endif
            MessageBox(NULL,"error decompressing image","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
            exit(0);
       }

       CD.bufferPos = (((m*60+s-2)*75)+f)/BZIP_BUFFER_SECTORS;
       CD.bufferPos *=2352*BZIP_BUFFER_SECTORS;
   }
   else if (CD.type == IndexZ)
   {
      unsigned int seekStart = CD.indexArray[((s-2+m*60)*75+f)];
      unsigned int seekEnd = CD.indexArray[((s-2+m*60)*75+f)+1];
      unsigned int destLen = seekEnd - seekStart;
      char tempBuffer[2352];

      rc = fseek(CD.cd, seekStart, SEEK_SET);
      rc = fread(tempBuffer, 1, destLen, CD.cd);

		uncompress(CD.buffer, &seekStart, tempBuffer, destLen);

       CD.bufferPos = (((m*60+s-2)*75)+f)*2352;
   }

   else
   {
   rc = fseek(CD.cd, CD.sector, SEEK_SET);
#ifdef SUPER_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " seek1 rc %d\n", rc);
   fclose(log);
#endif   

   rc = fread(CD.buffer, BUFFER_SIZE, sizeof(unsigned char), CD.cd);

#ifdef SUPER_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " seek2 rc %d\n", rc);
   fclose(log);
#endif   
   CD.bufferPos = CD.sector;
   
   }

}

// return the sector address - the buffer address + 12 bytes for offset.
unsigned char* getSector()
{
#ifdef SUPER_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, "getsector %u\n",CD.sector);
   fclose(log);
#endif

   if (CD.fullBuffer)
   {
      return CD.fullBuffer + CD.sector + 12;
   }
   else
   {
      return CD.buffer + (CD.sector - CD.bufferPos) + 12;
   }
}
