#include "cdreader.h"

#include "urarlib.h"

// llopen.c


// home of all the low level file opening functions



// my cue parsing function. it's messy, but i hope the comments help.
// NOTE: this will only parse playstation cd cue sheets correctly.
// any others will need to be modified to work correctly.  probably.

void readyImage()
{
   long end, size, blocks;

   CD.cd = fopen(CD.fileName, "rb");

   if (!CD.cd)
   {
      char str[350];
      sprintf(str, "Can't find file %s", CD.fileName);
#ifdef _WINDOWS
      MessageBox(NULL,str,"CDRMooby Exiting",MB_OK|MB_SETFOREGROUND);
#endif
      exit(0);
   }

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " readyimage opened %s \n", CD.fileName);
   fclose(log);
#endif   

   rc = fseek(CD.cd, 0, SEEK_END);

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " openbin1 rc %d\n", rc);
   fclose(log);
#endif   

   end = ftell(CD.cd);
   size = end;

   blocks = size / 2352;

   // put the track length info in the track list

   CD.tl[0].type = Mode2;
   CD.tl[0].num = 1;
   CD.tl[0].time[2] = blocks % 75;
   CD.tl[0].time[1] = ((blocks - CD.tl[0].time[2]) / 75) % 60;
   CD.tl[0].time[0] = (((blocks - CD.tl[0].time[2]) / 75) - CD.tl[0].time[1]) / 60;

   CD.tl[0].time[1] += 4;

   normalizeTime(CD.tl[0].time);

   rc = fseek(CD.cd, 0, SEEK_SET);

}

void parseCue(FILE* cue)
{
   
   Track t;
   char file[CHAR_LEN], id[CHAR_LEN], time[CHAR_LEN];
   int num, pregap;
   int min, sec, frac;
   int i;
   char c;   
   
   if (!cue) 
   {
#ifdef _WINDOWS
      MessageBox(NULL,"Can't open cue file","CDRMooby Exiting",MB_OK|MB_SETFOREGROUND);
#endif
      exit(0);
   }
   
   rc = fscanf(cue, "%s %s %s", &id, &file, &time);

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " opencue1 %s %s %s\n", file, id, time);
   fclose(log);
#endif

   file[strlen(file)-1] = '\0';

   strcpy(CD.fileName, file+1);

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " opened %s\n", CD.fileName);
   fclose(log);
#endif


   CD.numtracks = 0;

   pregap = 0;

   // parse mania...
   while (!feof(cue))
   {
      // based on the keyword (id), do something....
      fscanf(cue, "%s", &id);

      if (feof(cue)) break;

      // track signifies a new track
      if (strstr(id,"TRACK") != 0)
      {
         CD.numtracks++;

         // get the track number and mode info
         fscanf(cue, "%d %s", &num, &time);

         t.num = num;

         if (strstr(time,"MODE2/2352") != 0)
         {
            t.type = Mode2;
         }
         else if (strstr(time,"MODE1/2352") != 0)
         {
            t.type = Mode1;
         }
         else if (strstr(time,"AUDIO") != 0)
         {
            t.type = Audio;
         }
         else 
         {
            t.type = unknown;
         }
      }
      // if its a pregap, record the gap time
      else if (strstr(id,"PREGAP") != 0)
      {
         pregap = 1;

         fscanf(cue, "%2d%c%2d%c%2d", &min, &c, &sec, &c, &frac);
         t.time[0] = 0; //min;
         t.time[1] = 0; //sec;
         t.time[2] = 0; //frac;
      }
      // for index... is it index 0 or 1?
      else if (strstr(id,"INDEX") != 0)
      {
         fscanf(cue, "%d", &num);
         
         // index 0 is track start time
         if (num == 0)
         {
            // do nothing here =D
            fscanf(cue, "%2d%c%2d%c%2d", &min, &c, &sec, &c, &frac);
         }
         // index 1 is post-pregap time, i think... its also
         // the last thing you get before a new track so add it
         // to the track list
         else if (num == 1)
         {
            fscanf(cue, "%2d%c%2d%c%2d", &min, &c, &sec, &c, &frac);
            t.time[0] = min;
            t.time[1] = sec;
            t.time[2] = frac;

            // reallocate track list and add new track
            CD.tl = realloc(CD.tl, (CD.numtracks + 1) * sizeof(Track));
            CD.tl[CD.numtracks].time[0] = t.time[0];

            // the +2's are for correct timing of CD gaps.
            // there is an initial 2 sec offset, then 2 sec
            // gaps between every track
            CD.tl[CD.numtracks].time[1] = t.time[1]+2;
            if (CD.numtracks > 1)
            {
               CD.tl[CD.numtracks].time[1] += 2;
            }
            CD.tl[CD.numtracks].time[2] = t.time[2];
            CD.tl[CD.numtracks].num = t.num;
            CD.tl[CD.numtracks].type = t.type;

#ifdef MOOBY_DEBUG
            log = fopen(LOGFILE,"a");
            fprintf(log,"track %d %02d:%02d:%02d %02d:%02d:%02d\n",CD.numtracks,
                  (int)t.time[0],
                  (int)t.time[1],
                  (int)t.time[2],
                  (int)CD.tl[CD.numtracks].time[0],
                  (int)CD.tl[CD.numtracks].time[1],
                  (int)CD.tl[CD.numtracks].time[2]);
            fclose(log);
#endif

         }
      }
   }

   for (i = 1; i <= CD.numtracks; i++)
   {
      normalizeTime(CD.tl[i].time);
   }


   fclose(cue);
}

void openIndexBZ(const char* filename)
{
   char indexName[512];
   unsigned int indexArraySize = 0;
   long start;
   size_t cueLen;
   FILE* indexFile;
   strcpy(indexName,filename);
   lstrcat(indexName, ".index\0");


   indexFile = fopen(indexName, "rb");
   if (!indexFile)
   {
#ifdef _WINDOWS
         MessageBox(NULL,"No index file for bz file","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
#endif
         exit(0);

   }
   
   CD.indexArray = malloc(sizeof(unsigned int));
   fread(&start,1,sizeof(start),indexFile);

   while (!feof(indexFile))
   {
      indexArraySize++;
      CD.indexArray = realloc(CD.indexArray, indexArraySize * sizeof(unsigned int));
      CD.indexArray[indexArraySize-1] = start;
      fread(&start,1,sizeof(start),indexFile);
   }

   cueLen = CD.indexArray[indexArraySize-1];
   cueLen /= 2352;

   CD.numtracks = 1;
   CD.tl = realloc(CD.tl, (CD.numtracks + 1) * sizeof(Track));
   CD.tl[0].time[2] = cueLen % 75;
   CD.tl[0].time[1] = ((cueLen - CD.tl[0].time[2]) / 75) % 60;
   CD.tl[0].time[0] = (((cueLen - CD.tl[0].time[2]) / 75) - CD.tl[0].time[1]) / 60;
   CD.tl[0].time[1] += 4;

   normalizeTime(CD.tl[0].time);

   CD.tl[1].time[2] = 0;
   CD.tl[1].time[1] = 2;
   CD.tl[1].time[0] = 0;

   CD.cd = fopen(filename, "rb");

   if (!CD.cd)
   {
#ifdef _WINDOWS
         MessageBox(NULL,"Cant open the bz file","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
#endif
         exit(0);
   }
}

void openIndexZ(const char* filename)
{
   char indexName[512];
   unsigned int indexArraySize = 0;
   unsigned int start;
   size_t cueLen;
   unsigned short temp;
   FILE* indexFile;
   strcpy(indexName,filename);
   lstrcat(indexName, ".table\0");


   indexFile = fopen(indexName, "rb");
   if (!indexFile)
   {
#ifdef _WINDOWS
         MessageBox(NULL,"No index file for Z file","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
#endif
         exit(0);

   }
   
   CD.indexArray = malloc(sizeof(unsigned int));

   fread(&start,1,sizeof(start),indexFile);
   fread(&temp, 1, 2, indexFile);

   while (!feof(indexFile))
   {
      indexArraySize++;

      CD.indexArray = realloc(CD.indexArray, indexArraySize * sizeof(unsigned int));

      CD.indexArray[indexArraySize-1] = start;
      
      fread(&start,1,sizeof(start),indexFile);
      fread(&temp, 1, 2, indexFile);
   }

   indexArraySize++;

   CD.indexArray = realloc(CD.indexArray, indexArraySize * sizeof(unsigned int));

   CD.indexArray[indexArraySize-1] = start + temp;

   cueLen = CD.indexArray[indexArraySize-1];
   cueLen /= 2352;

   CD.numtracks = 1;
   CD.tl = realloc(CD.tl, (CD.numtracks + 1) * sizeof(Track));
   CD.tl[0].time[2] = cueLen % 75;
   CD.tl[0].time[1] = ((cueLen - CD.tl[0].time[2]) / 75) % 60;
   CD.tl[0].time[0] = (((cueLen - CD.tl[0].time[2]) / 75) - CD.tl[0].time[1]) / 60;
   CD.tl[0].time[1] += 4;

   normalizeTime(CD.tl[0].time);

   CD.tl[1].time[2] = 0;
   CD.tl[1].time[1] = 2;
   CD.tl[1].time[0] = 0;

   CD.cd = fopen(filename, "rb");

   if (!CD.cd)
   {
#ifdef _WINDOWS
         MessageBox(NULL,"Cant open the Z file","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
#endif
         exit(0);
   }
}

void openRar(const char* filename)
{
   
   ArchiveList_struct *list,*itr;
   char* cueBuffer = NULL;
   unsigned long cueLen;
   char* charpos;
   char* slashpos;
   char cueFileName[CHAR_LEN];

   slashpos = strrchr(filename, '\\');
   strcpy(cueFileName, slashpos + 1);
   charpos = strstr(cueFileName, ".rar");
   *(charpos+1) = 'c';
   *(charpos+2) = 'u';
   *(charpos+3) = 'e';

   if (urarlib_get(&cueBuffer, &cueLen, cueFileName,filename, NULL))
   {
      FILE* tmpcue = tmpfile();
      fwrite(cueBuffer, cueLen, sizeof(char), tmpcue);
      fseek(tmpcue, 0, SEEK_SET);
      parseCue(tmpcue);
      free(cueBuffer);
      CD.type = Cue;
   }
   else
   {
      CD.type = Bin;
      urarlib_list(filename, &list);
      if (list == NULL)
      {
#ifdef _WINDOWS
         MessageBox(NULL,"Invalid rar file","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
#endif
         exit(0);
      }
      strcpy(CD.fileName,list->item.Name);
      itr = list;
      while(itr != NULL)
      {
         itr = list->next;
         free(list->item.Name);
         free(list);
      }
   }

   if (!urarlib_get(&CD.fullBuffer, &cueLen,  CD.fileName,filename,NULL))
   {
      
#ifdef _WINDOWS
         MessageBox(NULL,"couldnt extract image from rar file","CDRMooby exiting",MB_OK|MB_SETFOREGROUND);
#endif
         exit(0);

   }
   
   cueLen /= 2352;
   CD.tl[0].time[2] = cueLen % 75;
   CD.tl[0].time[1] = ((cueLen - CD.tl[0].time[2]) / 75) % 60;
   CD.tl[0].time[0] = (((cueLen - CD.tl[0].time[2]) / 75) - CD.tl[0].time[1]) / 60;


   CD.tl[0].time[1] += 4;

   normalizeTime(CD.tl[0].time);

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
      fprintf(log, "rar cueparse %u %02d:%02d:%02d\n", cueLen,
         (int)CD.tl[0].time[0],
         (int)CD.tl[0].time[1],
         (int)CD.tl[0].time[2]);
   fclose(log);
#endif
}

void openCue(const char* filename)
{
#ifdef MOOBY_DEBUG
   int i;
#endif

   FILE* cue;
   // pregap is whether or not there is a pregap listing in the 
   // cue sheet

   // open cue sheet
   cue = fopen(filename, "r");
   if (!cue) 
   {
      char str[350];
      sprintf(str, "Can't find file %s", filename);
#ifdef _WINDOWS
      MessageBox(NULL,str,"CDRMooby Exiting",MB_OK);
#endif
      exit(0);
   }

   parseCue(cue);
   
   // get the file name from the cue sheet (remove the "'s).
   // Openbin will then add tl[0] with the full cd length

   readyImage();  

   CD.cdda = fopen(CD.fileName,"rb");


#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   for(i = 0; i <= CD.numtracks; i++)
   {
      fprintf(log, "cueparse %d %02d:%02d:%02d\n", i,
         (int)CD.tl[i].time[0],
         (int)CD.tl[i].time[1],
         (int)CD.tl[i].time[2]);
   }
   fclose(log);
#endif
}

// opens a binary cd image and calculates its length
void openBin(const char* filename)
{
   
#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " open bin filename %s\n", filename);
   fclose(log);
#endif  

   CD.numtracks = 1;

   CD.tl = realloc(CD.tl, (CD.numtracks + 1) * sizeof(Track));
   
   strcpy(CD.fileName, filename);

   readyImage();

   // there's only 1 track whose offset is 2 seconds on a bin image
   CD.tl[1].time[0] = 0;
   CD.tl[1].time[1] = 2;
   CD.tl[1].time[2] = 0;
}

// test if this is an iso-9660 image or a binary image
void openIso(const char* filename)
{
   long end;
   char temp[2352];

   CD.cd = fopen(filename, "rb");

   if (!CD.cd)
   {
      char str[350];
      sprintf(str, "Can't find file %s", filename);
#ifdef _WINDOWS
      MessageBox(NULL,str,"CDRMooby Exiting",MB_OK|MB_SETFOREGROUND);
#endif
      exit(0);
   }

   rc = fseek(CD.cd, 0, SEEK_END);

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   fprintf(log, " openiso1 rc %d\n", rc);
   fclose(log);
#endif   

   end = ftell(CD.cd);

   fseek(CD.cd,0,SEEK_SET);

   // is this really an iso or a bin with a .iso?
   fread(&temp, sizeof(unsigned char), 2352, CD.cd);

   if ((temp[12] != 0) && (temp[13] != 2) && (temp[14] != 0) &&
       ((end % 2352) != 0))
   {
#ifdef MOOBY_DEBUG
      log = fopen(LOGFILE,"a");
      fprintf(log, "really an iso\n");
      fclose(log);
#endif   

#ifdef _WINDOWS
      MessageBox(NULL,"ISO-9660 isn't supported by.  Sorry.","CDRMooby Exiting",MB_OK|MB_SETFOREGROUND);
#endif
      exit(0);
   }

   fclose(CD.cd);
   openBin(filename);
   return;
}

// add the rest of the subchannel data for the frames that have none or
// are incomplete
void finishSubchannelData()
{
   long i;
   for (i = 0; i < TOTAL_CD_LENGTH; i++)
   {
      long offset = i * BYTES_PER_SUBCHANNEL_FRAME;
      if (CD.subchannelData[offset] == 0)
      {
         CD.subchannelData[offset+12] = 0x41;
         CD.subchannelData[offset+13] = 0x01;
         CD.subchannelData[offset+14] = 0x01;
         CD.subchannelData[offset+15] = i % 75;
         CD.subchannelData[offset+16] = (i / 75) % 60;
         CD.subchannelData[offset+17] = (i / 75) / 60;
         CD.subchannelData[offset+18] = 0;
         CD.subchannelData[offset+19] = i % 75;
         CD.subchannelData[offset+20] = (i / 75) % 60;
         CD.subchannelData[offset+21] = (i / 75) / 60;
      }
         // one day if i ever get bored, i'll correctly calculate the track
         // relative address on this one
      if (CD.subchannelData[offset] == 2)
      {

      }
      CD.subchannelData[offset] = 0;
   }
}

// the type 3 subchannel data is wrong - i can't calculate the track info
// without a cue sheet, and even then i don't feel like adding all the calculation
// necessary to detect what track number it's supposed to be ;)
void processSBI()
{
   char header[4];
   long offset;

   fread(header, 1, sizeof(header), CD.subchannel);

   while (!feof(CD.subchannel))
   {
      fread(header, 1, sizeof(header), CD.subchannel);

      offset = ((header[0] * 60 + header[1]) * 75 + header[2]) * BYTES_PER_SUBCHANNEL_FRAME;

      CD.subchannelData[offset] = header[3];
      offset += 12;

      if (header[3] == 1)
      {
         fread(&CD.subchannelData[offset],1,10,CD.subchannel);
      }
      // this is definately wrong
      else if (header[3] == 2)
      {
         CD.subchannelData[offset] = 0x41;
         CD.subchannelData[offset+1] = 0x01;
         CD.subchannelData[offset+2] = 0x01;
         fread(&CD.subchannelData[offset+3],1,3,CD.subchannel);
         CD.subchannelData[offset+6] = 0;
         CD.subchannelData[offset+7] = header[0];
         CD.subchannelData[offset+8] = header[1];
         CD.subchannelData[offset+9] = header[2];
      }
      else if (header[3] == 3)
      {
         CD.subchannelData[offset] = 0x41;
         CD.subchannelData[offset+1] = 0x01;
         CD.subchannelData[offset+2] = 0x01;
         CD.subchannelData[offset+3] = header[0];
         CD.subchannelData[offset+4] = header[1];
         CD.subchannelData[offset+5] = header[2];
         CD.subchannelData[offset+6] = 0;
         fread(&CD.subchannelData[offset+7],1,3,CD.subchannel);
      }
   }
}

void processM3S()
{
   int offset = 3 * 60 * 75 * BYTES_PER_SUBCHANNEL_FRAME;

   while (!feof(CD.subchannel))
   {
      CD.subchannelData[offset] = 1;
      offset += 12;
      fread(&(CD.subchannelData[offset]), 1, 16, CD.subchannel);
      memset(&(CD.subchannelData[offset+10]),0,6);
      offset += (96 - 12);
   }
}


// new file types should be added here and in the CDOpen function
void newCD(const char * filename)
{
#ifdef MOOBY_DEBUG
   int i;
#endif
   char tempName[CHAR_LEN];
   char tempName2[CHAR_LEN];
   char *c;
   
   // place holder for full CD size
   CD.tl = malloc(sizeof(Track));

   playing = 0;
      // init for RAR
   if (CD.fullBuffer)
      free(CD.fullBuffer);
   CD.fullBuffer = NULL;
   CD.cd = NULL;
   CD.cdda = NULL;
   CD.numtracks = 0;
   CD.bcdFix = 0;

   memset(CD.subchannelData, 0, sizeof(CD.subchannelData));

   if (strstr(filename,".bz.index") != 0)
   {
      c = strstr(filename,".index");
      *c = '\0';
      CD.type = IndexBZ;
      openIndexBZ(filename);
   }
   else if (strstr(filename,".bz") != 0)
   {
      CD.type = IndexBZ;
      openIndexBZ(filename);
   }
   else if (strstr(filename,".Z.table") != 0)
   {
      c = strstr(filename,".table");
      *c = '\0';
      CD.type = IndexZ;
      openIndexZ(filename);
   }
   else if (strstr(filename,".Z") != 0)
   {
      CD.type = IndexZ;
      openIndexZ(filename);
   }
   else if (  (strstr(filename,".cue") != 0) ||
         (strstr(filename,".CUE") != 0) )
   {
      CD.type = Cue;
      openCue(filename);
   }
   else if (  (strstr(filename,".rar") != 0) ||
              (strstr(filename,".RAR") != 0) ) 
   {
      CD.type = Rar;
      openRar(filename);
   }
   else if ( (strstr(filename,".iso") != 0) ||
             (strstr(filename,".ISO") != 0) )
   {
      CD.type = Bin;
      openIso(filename);
   }
   else
   {
      CD.type = Bin;
      openBin(filename);
   }

   strcpy(tempName, filename);
   strcpy(tempName2, filename);
   c = strstr(tempName,".");
   *c = '\0';
   c = strstr(tempName2,".");
   *c = '\0';

   // try to open the subchannel files (if any)
   if (CD.subchannel = fopen(lstrcat(tempName,".SBI\0"),"rb"))
   {
      CD.subchannelType = SBI;
      processSBI();
   }
   else if (CD.subchannel = fopen(lstrcat(tempName2,".M3S\0"),"rb"))
   {
      CD.subchannelType = M3S;
      processM3S();
   }
   finishSubchannelData();

   CD.bufferPos = 0x7FFFFFFF;
   seekSector(0,2,0);

#ifdef MOOBY_DEBUG
   log = fopen(LOGFILE,"a");
   for(i = 0; i <= CD.numtracks; i++)
   {
      fprintf(log, "cdnew %d %02d:%02d:%02d\n", i,
         (int)CD.tl[i].time[0],
         (int)CD.tl[i].time[1],
         (int)CD.tl[i].time[2]);
   }
   fclose(log);
#endif

}
